"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const command_1 = require("@oclif/command");
const inquirer_1 = require("inquirer");
const path = require("path");
const json_schema_1 = require("../../../lib/src/generators/contract/json-schema");
const openapi2_1 = require("../../../lib/src/generators/contract/openapi2");
const openapi3_1 = require("../../../lib/src/generators/contract/openapi3");
const axios_client_1 = require("../../../lib/src/generators/typescript/axios-client");
const express_server_1 = require("../../../lib/src/generators/typescript/express-server");
const types_1 = require("../../../lib/src/generators/typescript/types");
const validators_1 = require("../../../lib/src/generators/typescript/validators");
const output_1 = require("../../../lib/src/io/output");
const parser_1 = require("../../../lib/src/parser");
const sortBy = require("lodash/sortBy");
class Generate extends command_1.Command {
    async run() {
        const { flags } = this.parse(Generate);
        let { api: apiPath, language, generator, out: outDir } = flags;
        const api = await parser_1.parsePath(apiPath);
        if (!generator) {
            generator = (await inquirer_1.prompt({
                name: "Generator",
                type: "list",
                choices: sortBy(Object.keys(generators))
            })).Generator;
        }
        if (!generators[generator]) {
            this.error(`No such generator ${generator}. Available generators:\n${availableGeneratorsList()}`);
            this.exit(1);
        }
        if (!language) {
            language = (await inquirer_1.prompt({
                name: "Language",
                type: "list",
                choices: sortBy(Object.keys(generators[generator]))
            })).Language;
        }
        if (!outDir) {
            outDir = (await inquirer_1.prompt({
                name: "Output destination",
                default: "."
            }))["Output destination"];
        }
        if (!generators[generator][language]) {
            const otherGenerators = Object.entries(generators)
                .filter(([name, languages]) => language in languages)
                .map(([name, languages]) => name);
            if (otherGenerators.length === 0) {
                this.error(`${language} is not supported by any generator. Available generators:\n${availableGeneratorsList()}`);
            }
            else {
                this.error(`${language} is not a supported language for this generator.\n\nThe generator ${generator} supports the following languages:\n${Object.keys(generators[generator])
                    .map(name => `- ${name}`)
                    .join("\n")}\n\nOther generators that produce ${language} are available: ${otherGenerators.join(", ")}`);
            }
            this.exit(1);
        }
        const generatedFiles = generators[generator][language](api);
        for (const [relativePath, content] of Object.entries(generatedFiles)) {
            output_1.outputFile(outDir, relativePath, content);
        }
        this.log(`Generated the following files:`);
        Object.keys(generatedFiles).forEach(relativePath => this.log(`- ${path.join(outDir, relativePath)}`));
    }
}
Generate.description = "Runs a generator on an API. Used to produce client libraries, server boilerplates and well-known API contract formats such as OpenAPI.";
Generate.examples = [
    `$ api generate --language typescript --generator axios-client --out src/
Generated the following files:
- src/types.ts
- src/validators.ts
- src/client.ts
`
];
Generate.flags = {
    help: command_1.flags.help({ char: "h" }),
    api: command_1.flags.string({
        required: true,
        char: "a",
        description: "Path to a TypeScript API definition"
    }),
    language: command_1.flags.string({
        char: "l",
        description: "Language to generate"
    }),
    generator: command_1.flags.string({
        char: "g",
        description: "Generator to run"
    }),
    out: command_1.flags.string({
        char: "o",
        description: "Directory in which to output generated files"
    })
};
exports.default = Generate;
function availableGeneratorsList() {
    return Object.entries(generators)
        .map(([name, languages]) => `- ${name} (${Object.keys(languages).join(", ")})`)
        .join("\n");
}
const generators = {
    "json-schema": {
        json: api => ({
            "types.json": json_schema_1.generateJsonSchema(api, "json")
        }),
        yaml: api => ({
            "types.yml": json_schema_1.generateJsonSchema(api, "yaml")
        })
    },
    openapi2: {
        json: api => ({
            "api.json": openapi2_1.generateOpenApiV2(api, "json")
        }),
        yaml: api => ({
            "api.yml": openapi2_1.generateOpenApiV2(api, "yaml")
        })
    },
    openapi3: {
        json: api => ({
            "api.json": openapi3_1.generateOpenApiV3(api, "json")
        }),
        yaml: api => ({
            "api.yml": openapi3_1.generateOpenApiV3(api, "yaml")
        })
    },
    "axios-client": {
        typescript: api => ({
            "types.ts": types_1.generateTypesSource(api),
            "validators.ts": validators_1.generateValidatorsSource(api),
            "client.ts": axios_client_1.generateAxiosClientSource(api)
        })
    },
    "express-server": {
        typescript: api => ({
            "types.ts": types_1.generateTypesSource(api),
            "validators.ts": validators_1.generateValidatorsSource(api),
            "server.ts": express_server_1.generateExpressServerSource(api),
            ...Object.entries(api.endpoints).reduce((acc, [endpointName, endpoint]) => {
                acc[`endpoints/${endpointName}.ts`] = express_server_1.generateEndpointHandlerSource(api, endpointName, endpoint);
                return acc;
            }, {})
        })
    }
};
