"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const YAML = require("js-yaml");
const assert_never_1 = require("../../assert-never");
const compact = require("lodash/compact");
function generateJsonSchema(api, format) {
    const contract = jsonSchema(api);
    switch (format) {
        case "json":
            return JSON.stringify(contract, null, 2);
        case "yaml":
            return YAML.safeDump(contract);
        default:
            throw assert_never_1.default(format);
    }
}
exports.generateJsonSchema = generateJsonSchema;
function jsonSchema(api) {
    return {
        $schema: "http://json-schema.org/draft-07/schema#",
        definitions: Object.entries(api.types).reduce((acc, [typeName, type]) => {
            acc[typeName] = jsonTypeSchema(type);
            return acc;
        }, {})
    };
}
exports.jsonSchema = jsonSchema;
function jsonTypeSchema(type) {
    switch (type.kind) {
        case "void":
        case "null":
            return {
                type: "null"
            };
        case "boolean":
            return {
                type: "boolean"
            };
        case "boolean-constant":
            return {
                type: "boolean",
                const: type.value
            };
        case "string":
            return {
                type: "string"
            };
        case "string-constant":
            return {
                type: "string",
                const: type.value
            };
        case "number":
            return {
                type: "number"
            };
        case "integer-constant":
            return {
                type: "integer",
                const: type.value
            };
        case "object":
            return Object.entries(type.properties).reduce((acc, [name, type]) => {
                if (type.kind === "optional") {
                    type = type.optional;
                }
                else {
                    acc.required.push(name);
                }
                const schemaType = jsonTypeSchema(type);
                if (schemaType) {
                    acc.properties[name] = schemaType;
                }
                return acc;
            }, {
                type: "object",
                properties: {},
                required: []
            });
        case "array":
            const itemsType = jsonTypeSchema(type.elements);
            if (!itemsType) {
                throw new Error(`Unsupported void array`);
            }
            return {
                type: "array",
                items: itemsType
            };
        case "optional":
            throw new Error(`Unsupported top-level optional type`);
        case "union":
            const types = type.types.map(t => jsonTypeSchema(t));
            const withoutNullTypes = compact(types);
            if (withoutNullTypes.length !== types.length) {
                throw new Error(`Unsupported void type in union`);
            }
            return {
                oneOf: withoutNullTypes
            };
        case "type-reference":
            return {
                $ref: `#/definitions/${type.typeName}`
            };
        default:
            throw assert_never_1.default(type);
    }
}
exports.jsonTypeSchema = jsonTypeSchema;
