"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ts = require("typescript");
const assert_never_1 = require("../../assert-never");
const ts_writer_1 = require("./ts-writer");
const types_1 = require("./types");
const flatten = require("lodash/flatten");
const compact = require("lodash/compact");
function generateValidatorsSource(api) {
    const typeNames = Object.keys(api.types);
    return ts_writer_1.outputTypeScriptSource([
        ...(typeNames.length > 0
            ? [
                ts.createImportDeclaration(
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, ts.createImportClause(
                /*name*/ undefined, ts.createNamedImports(typeNames.map(typeName => ts.createImportSpecifier(
                /*propertyName*/ undefined, ts.createIdentifier(typeName))))), ts.createStringLiteral("./types"))
            ]
            : []),
        ...flatten(Object.entries(api.endpoints).map(([endpointName, endpoint]) => [
            generateValidator(endpointPropertyTypeName(endpointName, "request"), endpoint.requestType),
            ...compact(endpoint.path.map(pathComponent => pathComponent.kind === "dynamic"
                ? generateValidator(endpointPropertyTypeName(endpointName, "param", pathComponent.name), pathComponent.type)
                : null)),
            ...Object.entries(endpoint.headers).map(([headerName, header]) => generateValidator(endpointPropertyTypeName(endpointName, "header", headerName), header.type)),
            generateValidator(endpointPropertyTypeName(endpointName, "response"), endpoint.responseType),
            generateValidator(endpointPropertyTypeName(endpointName, "genericError"), endpoint.genericErrorType),
            ...Object.entries(endpoint.specificErrorTypes).map(([name, specificError]) => generateValidator(endpointPropertyTypeName(endpointName, `specificError`, name), specificError.type))
        ])),
        ...Object.entries(api.types).map(([typeName, type]) => generateValidator(typeName, type, typeName))
    ]);
}
exports.generateValidatorsSource = generateValidatorsSource;
function generateValidator(typeNickname, type, typeName) {
    const parameterName = ts.createIdentifier("value");
    return ts.createFunctionDeclaration(
    /*decorators*/ undefined, 
    /*modifiers*/ [ts.createToken(ts.SyntaxKind.ExportKeyword)], 
    /*asteriskToken*/ undefined, ts.createIdentifier(validatorName(typeNickname)), 
    /*typeParameters*/ undefined, [
        ts.createParameter(
        /*decorators*/ undefined, 
        /*modifiers*/ undefined, 
        /*dotDotDotToken*/ undefined, parameterName, 
        /*questionToken*/ undefined, ts.createKeywordTypeNode(ts.SyntaxKind.AnyKeyword))
    ], 
    /*returnType*/ ts.createTypePredicateNode(parameterName, typeName
        ? ts.createTypeReferenceNode(typeName, /*typeArguments*/ undefined)
        : types_1.typeNode(type)), ts.createBlock([ts.createReturn(validator(type, parameterName))], 
    /*multiline*/ true));
}
function validator(type, parameter) {
    switch (type.kind) {
        case "void":
            return voidValidator(parameter);
        case "null":
            return nullValidator(parameter);
        case "boolean":
            return booleanValidator(parameter);
        case "boolean-constant":
            return booleanConstantValidator(type.value, parameter);
        case "string":
            return stringValidator(parameter);
        case "string-constant":
            return stringConstantValidator(type.value, parameter);
        case "number":
            return numberValidator(parameter);
        case "integer-constant":
            return integerConstantValidator(type.value, parameter);
        case "object":
            return objectValidator(type, parameter);
        case "array":
            return arrayValidator(type, parameter);
        case "optional":
            return optionalValidator(type, parameter);
        case "union":
            return unionValidator(type, parameter);
        case "type-reference":
            return typeReferenceValidator(type, parameter);
        default:
            throw assert_never_1.default(type);
    }
}
function voidValidator(parameter) {
    return ts.createStrictEquality(parameter, ts.createIdentifier("undefined"));
}
function nullValidator(parameter) {
    return ts.createStrictEquality(parameter, ts.createIdentifier("null"));
}
function booleanValidator(parameter) {
    return ts.createStrictEquality(ts.createTypeOf(parameter), ts.createStringLiteral("boolean"));
}
function stringValidator(parameter) {
    return ts.createStrictEquality(ts.createTypeOf(parameter), ts.createStringLiteral("string"));
}
function numberValidator(parameter) {
    return ts.createStrictEquality(ts.createTypeOf(parameter), ts.createStringLiteral("number"));
}
function notNullObjectValidator(parameter) {
    return ts.createLogicalAnd(ts.createLogicalNot(nullValidator(parameter)), ts.createStrictEquality(ts.createTypeOf(parameter), ts.createStringLiteral("object")));
}
function booleanConstantValidator(value, parameter) {
    return ts.createStrictEquality(parameter, ts.createIdentifier(value ? "true" : "false"));
}
function stringConstantValidator(value, parameter) {
    return ts.createStrictEquality(parameter, ts.createStringLiteral(value));
}
function integerConstantValidator(value, parameter) {
    return ts.createStrictEquality(parameter, ts.createNumericLiteral(value.toString(10)));
}
function objectValidator(type, parameter) {
    let expression = notNullObjectValidator(parameter);
    for (const [propertyName, propertyType] of Object.entries(type.properties)) {
        expression = ts.createLogicalAnd(expression, validator(propertyType, ts.createElementAccess(parameter, ts.createStringLiteral(propertyName))));
    }
    return expression;
}
function arrayValidator(type, parameter) {
    const accArgument = ts.createIdentifier("acc");
    const currArgument = ts.createIdentifier("curr");
    return ts.createLogicalAnd(ts.createBinary(parameter, ts.SyntaxKind.InstanceOfKeyword, ts.createIdentifier("Array")), ts.createCall(ts.createPropertyAccess(parameter, "reduce"), 
    /*typeArguments*/ undefined, [
        ts.createArrowFunction(
        /*modifiers*/ undefined, 
        /*typeParameters*/ undefined, [
            ts.createParameter(
            /*decorators*/ undefined, 
            /*modifiers*/ undefined, 
            /*dotDotDotToken*/ undefined, accArgument),
            ts.createParameter(
            /*decorators*/ undefined, 
            /*modifiers*/ undefined, 
            /*dotDotDotToken*/ undefined, currArgument)
        ], 
        /*type*/ undefined, 
        /*equalsGreaterThanToken*/ undefined, ts.createLogicalAnd(accArgument, validator(type.elements, currArgument))),
        ts.createIdentifier("true")
    ]));
}
function optionalValidator(type, parameter) {
    return ts.createLogicalOr(voidValidator(parameter), validator(type.optional, parameter));
}
function unionValidator(type, parameter) {
    let expression = validator(type.types[0], parameter);
    for (let i = 1; i < type.types.length; i++) {
        expression = ts.createLogicalOr(expression, validator(type.types[i], parameter));
    }
    return expression;
}
function typeReferenceValidator(type, parameter) {
    return ts.createCall(ts.createIdentifier(validatorName(type.typeName)), 
    /*typeArguments*/ undefined, [parameter]);
}
function endpointPropertyTypeName(endpointName, property, suffix = "") {
    switch (property) {
        case "param":
        case "header":
        case "specificError":
            if (!suffix) {
                throw new Error(`Unexpected ${property} without a suffix`);
            }
            break;
        default:
            // No suffix required.
            if (suffix) {
                throw new Error(`Unexpected ${property} with a suffix`);
            }
    }
    return `${endpointName}_${property}${capitaliseFirstLetter(suffix)}`;
}
exports.endpointPropertyTypeName = endpointPropertyTypeName;
function validatorName(typeName) {
    return `validate${capitaliseFirstLetter(typeName)}`;
}
exports.validatorName = validatorName;
function capitaliseFirstLetter(name) {
    if (name.length === 0) {
        return name;
    }
    return `${name[0].toUpperCase()}${name.substr(1)}`;
}
function validateStatement(variable, validatorName, errorMessage) {
    return ts.createIf(ts.createLogicalNot(ts.createCall(ts.createPropertyAccess(ts.createIdentifier("validators"), validatorName), 
    /*typeArguments*/ undefined, [variable])), ts.createBlock([
        ts.createThrow(ts.createNew(ts.createIdentifier("Error"), 
        /*typeArguments*/ undefined, [
            ts.createTemplateExpression(ts.createTemplateHead(`${errorMessage}: `), [
                ts.createTemplateSpan(ts.createCall(ts.createPropertyAccess(ts.createIdentifier("JSON"), "stringify"), 
                /*typeArguments*/ undefined, [
                    variable,
                    ts.createIdentifier("null"),
                    ts.createNumericLiteral("2")
                ]), ts.createTemplateTail(""))
            ])
        ]))
    ], 
    /*multiLine*/ true));
}
exports.validateStatement = validateStatement;
