"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const assert_never_1 = require("../../assert-never");
const compact = require("lodash/compact");
function rejectVoidOpenApi3SchemaType(type, errorMessage) {
    const schemaType = openApi3TypeSchema(type);
    if (!schemaType) {
        throw new Error(errorMessage);
    }
    return schemaType;
}
exports.rejectVoidOpenApi3SchemaType = rejectVoidOpenApi3SchemaType;
function openApiV3ContentTypeSchema(contentType, type) {
    switch (contentType) {
        case "application/json":
            return {
                content: {
                    "application/json": {
                        schema: openApi3TypeSchema(type)
                    }
                }
            };
        case "text/html":
            return {
                content: {
                    "text/html": {
                        schema: {
                            type: "string"
                        }
                    }
                }
            };
        default:
            throw assert_never_1.default(contentType);
    }
}
exports.openApiV3ContentTypeSchema = openApiV3ContentTypeSchema;
function openApi3TypeSchema(type) {
    switch (type.kind) {
        case "void":
            return null;
        case "null":
            return {
                nullable: true
            };
        case "boolean":
            return {
                type: "boolean"
            };
        case "boolean-constant":
            return {
                type: "boolean",
                enum: [type.value]
            };
        case "string":
            return {
                type: "string"
            };
        case "string-constant":
            return {
                type: "string",
                enum: [type.value]
            };
        case "number":
            return {
                type: "number"
            };
        case "integer-constant":
            return {
                type: "integer",
                enum: [type.value]
            };
        case "object":
            return Object.entries(type.properties).reduce((acc, [name, type]) => {
                if (type.kind === "optional") {
                    type = type.optional;
                }
                else {
                    acc.required.push(name);
                }
                const schemaType = openApi3TypeSchema(type);
                if (schemaType) {
                    acc.properties[name] = schemaType;
                }
                return acc;
            }, {
                type: "object",
                properties: {},
                required: []
            });
        case "array":
            const itemsType = openApi3TypeSchema(type.elements);
            if (!itemsType) {
                throw new Error(`Unsupported void array`);
            }
            return {
                type: "array",
                items: itemsType
            };
        case "optional":
            throw new Error(`Unsupported top-level optional type`);
        case "union":
            const types = type.types.map(t => openApi3TypeSchema(t));
            const withoutNullTypes = compact(types);
            if (withoutNullTypes.length !== types.length) {
                throw new Error(`Unsupported void type in union`);
            }
            return {
                oneOf: withoutNullTypes
            };
        case "type-reference":
            return {
                $ref: `#/components/schemas/${type.typeName}`
            };
        default:
            throw assert_never_1.default(type);
    }
}
exports.openApi3TypeSchema = openApi3TypeSchema;
