"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ts = require("typescript");
const assert_never_1 = require("../../assert-never");
const models_1 = require("../../models");
const validator_1 = require("../../validator");
const ts_writer_1 = require("./ts-writer");
const types_1 = require("./types");
const validators_1 = require("./validators");
const compact = require("lodash/compact");
const IMPORTED_AXIOS_NAME = "axios";
function generateAxiosClientSource(api) {
    const typeNames = Object.keys(api.types);
    return ts_writer_1.outputTypeScriptSource([
        ts.createImportDeclaration(
        /*decorators*/ undefined, 
        /*modifiers*/ undefined, ts.createImportClause(ts.createIdentifier(IMPORTED_AXIOS_NAME), 
        /*namedBindings*/ undefined), ts.createStringLiteral("axios")),
        ...(typeNames.length > 0
            ? [
                ts.createImportDeclaration(
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, ts.createImportClause(
                /*name*/ undefined, ts.createNamedImports(typeNames.map(typeName => ts.createImportSpecifier(
                /*propertyName*/ undefined, ts.createIdentifier(typeName))))), ts.createStringLiteral("./types"))
            ]
            : []),
        ts.createImportDeclaration(
        /*decorators*/ undefined, 
        /*modifiers*/ undefined, ts.createImportClause(
        /*name*/ undefined, ts.createNamespaceImport(ts.createIdentifier("validators"))), ts.createStringLiteral("./validators")),
        ...Object.entries(api.endpoints).map(([endpointName, endpoint]) => generateEndpointFunction(api, endpointName, endpoint))
    ]);
}
exports.generateAxiosClientSource = generateAxiosClientSource;
const REQUEST_PARAMETER = "request";
function generateEndpointFunction(api, endpointName, endpoint) {
    const includeRequest = !validator_1.isVoid(api, endpoint.requestType);
    return ts.createFunctionDeclaration(
    /*decorators*/ undefined, [
        ts.createToken(ts.SyntaxKind.ExportKeyword),
        ts.createToken(ts.SyntaxKind.AsyncKeyword)
    ], 
    /*asteriskToken*/ undefined, endpointName, 
    /*typeParameters*/ undefined, [
        ...(includeRequest
            ? [
                ts.createParameter(
                /*decorators*/ undefined, 
                /*modifiers*/ undefined, 
                /*dotDotDotToken*/ undefined, REQUEST_PARAMETER, 
                /*questionToken*/ undefined, types_1.typeNode(endpoint.requestType))
            ]
            : []),
        ...compact(endpoint.path.map(pathComponent => pathComponent.kind === "dynamic"
            ? ts.createParameter(
            /*decorators*/ undefined, 
            /*modifiers*/ undefined, 
            /*dotDotDotToken*/ undefined, pathComponent.name, 
            /*questionToken*/ undefined, types_1.typeNode(pathComponent.type))
            : null)),
        ...Object.entries(endpoint.headers).map(([headerName, header]) => ts.createParameter(
        /*decorators*/ undefined, 
        /*modifiers*/ undefined, 
        /*dotDotDotToken*/ undefined, headerName, 
        /*questionToken*/ undefined, types_1.typeNode(header.type)))
    ], types_1.promiseTypeNode(models_1.unionType(...generateReturnTypes(endpoint))), generateEndpointBody(endpointName, endpoint, includeRequest));
}
function generateReturnTypes(endpoint) {
    return [
        models_1.objectType({
            kind: models_1.stringConstant("success"),
            data: endpoint.responseType
        }),
        models_1.objectType({
            kind: models_1.stringConstant("unknown-error"),
            data: endpoint.genericErrorType
        }),
        ...Object.entries(endpoint.specificErrorTypes).map(([name, specificError]) => models_1.objectType({
            kind: models_1.stringConstant(name),
            data: specificError.type
        }))
    ];
}
function generateEndpointBody(endpointName, endpoint, includeRequest) {
    return ts.createBlock([
        ...(includeRequest
            ? [generateRequestValidation(endpointName, endpoint)]
            : []),
        ...compact(endpoint.path.map(pathComponent => pathComponent.kind === "dynamic"
            ? generatePathParameterValidation(endpointName, pathComponent)
            : null)),
        ...Object.keys(endpoint.headers).map(headerName => generateHeaderValidation(endpointName, headerName)),
        generateAxiosCall(endpoint, includeRequest),
        generateSwitchStatus(endpointName, endpoint)
    ], 
    /*multiLine*/ true);
}
function generateRequestValidation(endpointName, endpoint) {
    return validators_1.validateStatement(ts.createIdentifier(REQUEST_PARAMETER), validators_1.validatorName(validators_1.endpointPropertyTypeName(endpointName, REQUEST_PARAMETER)), "Invalid request");
}
function generateHeaderValidation(endpointName, headerName) {
    return validators_1.validateStatement(ts.createIdentifier(headerName), validators_1.validatorName(validators_1.endpointPropertyTypeName(endpointName, "header", headerName)), `Invalid parameter ${headerName}`);
}
function generatePathParameterValidation(endpointName, pathComponent) {
    return validators_1.validateStatement(ts.createIdentifier(pathComponent.name), validators_1.validatorName(validators_1.endpointPropertyTypeName(endpointName, "param", pathComponent.name)), `Invalid parameter ${pathComponent.name}`);
}
const RESPONSE_NAME = "response";
const RESPONSE_STATUS_CODE = ts.createPropertyAccess(ts.createIdentifier(RESPONSE_NAME), "status");
const RESPONSE_DATA = ts.createPropertyAccess(ts.createIdentifier(RESPONSE_NAME), "data");
function generateAxiosCall(endpoint, includeRequest) {
    return ts.createVariableStatement(
    /*modifiers*/ undefined, ts.createVariableDeclarationList([
        ts.createVariableDeclaration(RESPONSE_NAME, 
        /*type*/ undefined, ts.createAwait(ts.createCall(ts.createIdentifier(IMPORTED_AXIOS_NAME), 
        /*typeArguments*/ undefined, [
            ts.createObjectLiteral([
                ts.createPropertyAssignment("url", generatePath(endpoint.path)),
                ts.createPropertyAssignment("method", ts.createStringLiteral(endpoint.method)),
                ts.createPropertyAssignment("responseType", ts.createStringLiteral("json")),
                ts.createPropertyAssignment("headers", ts.createObjectLiteral(Object.entries(endpoint.headers).map(([headerName, header]) => ts.createPropertyAssignment(header.headerFieldName, ts.createIdentifier(headerName))))),
                ...(includeRequest
                    ? [
                        ts.createPropertyAssignment("data", ts.createIdentifier(REQUEST_PARAMETER))
                    ]
                    : []),
                ts.createPropertyAssignment("validateStatus", ts.createArrowFunction(
                /*modifiers*/ undefined, 
                /*typeParameters*/ undefined, [], 
                /*type*/ undefined, 
                /*equalsGreaterThanToken*/ undefined, ts.createLiteral(true)))
            ], 
            /*multiLine*/ true)
        ])))
    ], ts.NodeFlags.Const));
}
function generatePath(path) {
    let result = null;
    for (const pathComponent of path) {
        let append;
        switch (pathComponent.kind) {
            case "static":
                append = ts.createStringLiteral(pathComponent.content);
                break;
            case "dynamic":
                append = ts.createIdentifier(pathComponent.name);
                break;
            default:
                throw assert_never_1.default(pathComponent);
        }
        if (!result) {
            result = append;
        }
        else {
            result = ts.createBinary(result, ts.SyntaxKind.PlusToken, append);
        }
    }
    if (!result) {
        result = ts.createStringLiteral("");
    }
    return result;
}
function generateSwitchStatus(endpointName, endpoint) {
    return ts.createSwitch(RESPONSE_STATUS_CODE, ts.createCaseBlock([
        ...Object.entries(endpoint.specificErrorTypes).map(([name, specificError]) => ts.createCaseClause(ts.createNumericLiteral(specificError.statusCode.toString(10)), [
            validators_1.validateStatement(RESPONSE_DATA, validators_1.validatorName(validators_1.endpointPropertyTypeName(endpointName, "specificError", name)), `Invalid response for status code ${specificError.statusCode}`),
            ts.createReturn(ts.createObjectLiteral([
                ts.createPropertyAssignment("kind", ts.createStringLiteral(name)),
                ts.createPropertyAssignment("data", RESPONSE_DATA)
            ], 
            /*multiLine*/ true))
        ])),
        ts.createDefaultClause([
            ts.createIf(ts.createLogicalAnd(ts.createBinary(RESPONSE_STATUS_CODE, ts.SyntaxKind.GreaterThanEqualsToken, ts.createNumericLiteral("200")), ts.createBinary(RESPONSE_STATUS_CODE, ts.SyntaxKind.LessThanToken, ts.createNumericLiteral("300"))), ts.createBlock([
                validators_1.validateStatement(RESPONSE_DATA, validators_1.validatorName(validators_1.endpointPropertyTypeName(endpointName, "response")), `Invalid response for successful status code`),
                ts.createReturn(ts.createObjectLiteral([
                    ts.createPropertyAssignment("kind", ts.createStringLiteral("success")),
                    ts.createPropertyAssignment("data", RESPONSE_DATA)
                ], 
                /*multiLine*/ true))
            ], 
            /*multiLine*/ true), ts.createBlock([
                validators_1.validateStatement(RESPONSE_DATA, validators_1.validatorName(validators_1.endpointPropertyTypeName(endpointName, "genericError")), `Invalid response for unknown error`),
                ts.createReturn(ts.createObjectLiteral([
                    ts.createPropertyAssignment("kind", ts.createStringLiteral("unknown-error")),
                    ts.createPropertyAssignment("data", RESPONSE_DATA)
                ], 
                /*multiLine*/ true))
            ], 
            /*multiLine*/ true))
        ])
    ]));
}
